# Copyright (c) 2023-2025 Fumito Hamamura <fumito.ham@gmail.com>

# This library is free software: you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation version 3.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library.  If not, see <http://www.gnu.org/licenses/>.

import sys
import os
import textwrap
import pathlib
import shutil
import runpy
import ast
import argparse
import logging
import subprocess
from typing import IO, TYPE_CHECKING, Sequence, Optional, Tuple

from modelx_cython.consts import MX_MODEL_MOD, MX_SPACE_MOD, MX_SYS_MOD
from modelx_cython.config import TransSpec
from modelx_cython.tracer import trace_calls, MxCallTraceLogger, MxCodeFilter
from modelx_cython.builder import ModuleInfo
from modelx_cython.parser import ModuleVisitor
from modelx_cython.transformer import ModuleTransformer, PXDGenerator


def increment_backups(
        base_path: pathlib.Path,
        max_backups=1, nth=0):

    postfix = "_BAK" + str(nth) if nth else ""
    backup_path = pathlib.Path(str(base_path) + postfix)
    if backup_path.exists():
        if nth == max_backups:
            if backup_path.is_dir():
                shutil.rmtree(backup_path)
            elif backup_path.is_file():
                backup_path.unlink()
            else:
                raise ValueError("cannot remove '%s'" % str(backup_path))
        else:
            increment_backups(base_path, max_backups, nth + 1)
            next_backup = pathlib.Path(str(base_path) + "_BAK" + str(nth + 1))
            backup_path.rename(next_backup)


def run_sample(model_path: pathlib.Path, sample_path: str, new_model_name: str = None) -> MxCallTraceLogger:

    module: str = model_path.name
    try:
        module_path = str(model_path.parent)
        sys.path.insert(0, module_path)

        logger = MxCallTraceLogger(module=module, new_model_name=new_model_name)
        with trace_calls(
            module=module,
            logger=logger,
            max_typed_dict_size=0,
            code_filter= MxCodeFilter()):
            runpy.run_path(sample_path, run_name="__main__")

    finally:
        assert sys.path.pop(0) == module_path

    return logger


class HandlerError(Exception):
    pass


def main_handler(args: argparse.Namespace, stdout: IO[str], stderr: IO[str]) -> int:

    orig_path = pathlib.Path(args.model_path).resolve()
    work_dir = orig_path.parent
    model_name = orig_path.name + "_cy"
    model_path = work_dir / model_name
    setup_file = pathlib.Path(args.setup) if args.setup else work_dir / "setup.py"

    if not args.compile_only:
        increment_backups(model_path)
        shutil.copytree(orig_path, model_path)
        shutil.copy(pathlib.Path(__file__).parent / (MX_SYS_MOD + ".pxd"), model_path)

        logger = run_sample(orig_path, args.sample, new_model_name=model_name)
        if args.no_spec:
            d = {}
        else:
            try:
                d = ast.literal_eval(pathlib.Path(args.spec).read_text())
            except FileNotFoundError as e:
                raise FileNotFoundError(f"{e}. Add '--no-spec' to omit the spec file.") from e

        spec = TransSpec(d)
        rel_model_path = model_path.relative_to(model_path.parent)

        modules = [rel_model_path / (MX_SYS_MOD + ".py")]
        for m in logger.modules:
            subs = m.split(".")
            assert subs.pop(0) == model_path.name
            assert subs[-1] in [MX_MODEL_MOD, MX_SPACE_MOD]
            pxd_path = subs.copy()
            subs[-1] = subs[-1] + ".py"
            pxd_path[-1] = pxd_path[-1] + ".pxd"
            abs_src_path = model_path / "/".join(subs)
            rel_src_path = rel_model_path / "/".join(subs)
            abs_pxd_path = model_path / "/".join(pxd_path)
            abs_init_path = model_path / "/".join(subs[:-1] + ["__init__.pxd"])
            source = abs_src_path.read_text()
            visitor = ModuleVisitor(module=m, source=source)
            module_info = ModuleInfo(m, visitor, logger, spec)
            trans = ModuleTransformer(source, module_info)
            pxd = PXDGenerator(module_info)

            abs_src_path.write_text(trans.transformed.code)
            abs_pxd_path.write_text(pxd.code)
            abs_init_path.write_text("from . cimport _mx_classes")
            modules.append(rel_src_path)

        create_setup(model_name, modules=modules, setup_file=setup_file)

    if args.translate_only:
        return 0
    else:
        return compile_main(work_dir, setup_file)


# Mapping from string names to logging levels
LOG_LEVELS = {
    'notset': logging.NOTSET,
    'debug': logging.DEBUG,
    'info': logging.INFO,
    'warning': logging.WARNING,
    'error': logging.ERROR,
    'critical': logging.CRITICAL
}


# Custom type function for argparse
def parse_log_level(value):
    try:
        # Try interpreting as an integer
        level = int(value)
        if level in LOG_LEVELS.values():
            return level
        raise ValueError
    except ValueError:
        # Try interpreting as a string key
        level_name = value.lower()
        if level_name in LOG_LEVELS:
            return LOG_LEVELS[level_name]
        raise argparse.ArgumentTypeError(f"Invalid log level: {value}")


def compile_main(work_dir: pathlib.Path, setup_file: pathlib.Path) -> int:

    env = os.environ.copy()
    env["PYTHONPATH"] = str(work_dir) + os.pathsep + env.get("PYTHONPATH", "")
    cmd = subprocess.run([sys.executable, str(setup_file), "build_ext", "--inplace"],
                         env=env, cwd=str(work_dir))
    return cmd.returncode


def main(argv: Sequence[str], stdout: IO[str], stderr: IO[str]) -> int:

    parser = argparse.ArgumentParser(
        description="Translate an exported modelx model into Cython and compile it.",
    )

    parser.add_argument(
        "model_path",
        type=str,
        help="Path to an exported modelx model to translate into Cython",
    )

    parser.add_argument(
        "--sample",
        type=str,
        default="sample.py",
        help=(
            "Path to a sample file to run for collecting type information (default: sample.py)"
        )
    )

    spec_group = parser.add_mutually_exclusive_group()

    spec_group.add_argument(
        "--spec",
        type=str,
        default="spec.py",
        help=(
            "Path to a spec file for setting parameters (default: spec.py)"
        )
    )

    spec_group.add_argument(
        "--no-spec",
        action="store_true",
        default=False,
        help="Skip the spec file (default: False)"
    )

    parser.add_argument(
        "--setup",
        type=str,
        default="",
        help=(
            "Path to a setup file for Cython (default: setup.py)"
        )
    )

    task_group = parser.add_mutually_exclusive_group()

    task_group.add_argument(
        "--translate-only",
        action="store_true",
        default=False,
        help="Perform translation only (default: False)",
    )

    task_group.add_argument(
        "--compile-only",
        action="store_true",
        default=False,
        help="Perform compilation only (default: False)",
    )

    parser.add_argument(
        '--log-level',
        default=logging.WARNING,
        type=parse_log_level,
        help='Logging level: NOTSET(0), DEBUG(10), INFO(20), WARNING(30), ERROR(40), CRITICAL(50) (default: WARNING)'
    )
    args = parser.parse_args(argv)

    # Configure logging
    logging.basicConfig(level=args.log_level)

    return main_handler(args, stdout, stderr)


def create_setup(model_name: str, modules: Sequence[str], setup_file: pathlib.Path):

    modules_str = textwrap.indent(",\n".join(
        ['"' + s.as_posix() + '"' for s in modules]
    ), " " * 8)

    setup_script = textwrap.dedent("""\
    import sys
    from setuptools import setup
    from Cython.Build import cythonize

    setup(
        name="{model_name}",
        ext_modules=cythonize([
    {modules_str}
            ],
            annotate=True
        )
    )
    """)

    setup_file.write_text(
        setup_script.format(
            model_name=model_name,
            modules_str=modules_str))


def entry_point_main():
    """Wrapper for main() for setuptools console_script entry point."""
    # Since monkeytype needs to import the user's code (and possibly config
    # code), the user's code must be on the Python path. But when running the
    # CLI script, it won't be. So we add the current working directory to the
    # Python path ourselves.
    # sys.path.insert(0, os.getcwd())
    sys.exit(main(sys.argv[1:], sys.stdout, sys.stderr))




# Copyright (c) 2023-2025 Fumito Hamamura <fumito.ham@gmail.com>

# This library is free software: you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation version 3.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library.  If not, see <http://www.gnu.org/licenses/>.



version_info = (0, 0, 8)
__version__ = ".".join([str(x) for x in version_info])



# Copyright (c) 2023-2025 Fumito Hamamura <fumito.ham@gmail.com>

# This library is free software: you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation version 3.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library.  If not, see <http://www.gnu.org/licenses/>.

# Portions of this source file are derived from MonkeyType, 
# a Python library for generating static type annotations from runtime types.
# Refer to the original license and copyright included with this library.


import sys
import pathlib
import random
import numbers
import itertools
from dataclasses import dataclass
from contextlib import contextmanager
from types import FrameType
from typing import Any, Mapping, Iterator, Sequence, Optional, Dict, List
import logging

import numpy as np

from modelx_cython.monkeytype_tracing import (
    CallTraceLogger,
    get_func,
    CodeFilter,
    CallTrace,
    CallTracer,
    RETURN_VALUE_OPCODE,
    YIELD_VALUE_OPCODE,
    SUPPORTED_EVENTS,
    EVENT_CALL,
    EVENT_RETURN,
)

from modelx_cython.consts import (
    MX_ASSIGN_REFS,
    FORMULA_PREF,
    is_user_defined,
    MX_SELF,
    MX_MODEL_MOD,
    MX_SPACE_MOD,
    MX_SYS_MOD,
    BASE_MODEL,
    SPACE_PREF,
    SPACE_PARAMS
)

if (3, 12) <= sys.version_info < (3, 14):
    import opcode
    RETURN_CONST_OPCODE = opcode.opmap["RETURN_CONST"]

_logger = logging.getLogger(__name__)


@dataclass
class ReturnTypeInfo:
    value_type: type
    is_array: bool = False
    ndim: int = 0


class RuntimeCellsInfo:     # TODO: Create base class RuntimeBaseMemberInfo
    name: str
    module: str
    arg_types: Dict[str, type]  # without self
    max_args: Dict[str, int]
    ret_type: ReturnTypeInfo

    def __init__(self, traces: Sequence[CallTrace]) -> None:
        self.fqname = traces[0].funcname
        self.name = traces[0].func.__name__
        self.module = traces[0].func.__module__
        self.arg_types = {}
        self.max_args = {}
        self._init_arg_types(traces)
        self.ret_type = self._init_ret_type(traces)

    def has_args(self):
        return bool(len(self.arg_types))

    def _init_arg_types(self, traces):
        arg_type_val: Dict[str, dict[type, Any]] = {}

        for trace in traces:

            for arg, val in itertools.islice(
                trace.arg_vals.items(), 1, None
            ):  # remove self
                tp = type(val)
                types = arg_type_val.setdefault(arg, {})
                if tp not in types:
                    types[tp] = val
                elif issubclass(tp, numbers.Integral):
                    if val > types[tp]:
                        types[tp] = val

        for arg, type_val in arg_type_val.items():

            if len(type_val) == 1:
                for tp, val in type_val.items():
                    self.arg_types[arg] = tp
                    if issubclass(tp, numbers.Integral):
                        self.max_args[arg] = val
                    break
            elif all(issubclass(tp, numbers.Integral) for tp in type_val.keys()):
                self.arg_types[arg] = numbers.Integral
                self.max_args[arg] = max(v for v in type_val.values())

            elif all(issubclass(tp, str) for tp in type_val.keys()):
                self.arg_types[arg] = str

            else:
                if (any(issubclass(tp, numbers.Integral) for tp in type_val.keys()) and
                        any(not issubclass(tp, numbers.Integral) for tp in type_val.keys())):
                    msg = ', '.join(k.__name__ + " " + str(v) for k, v in type_val.items())
                    _logger.info(f"varying types given to argument '{arg}' in {self.fqname}: {msg}")

                self.arg_types[arg] = object

    def _init_ret_type(self, traces):
        last_tp = None
        last_args = None
        has_last = False
        was_dtype_logged = False
        was_ndim_logged = False
        was_vtype_logged = False

        def get_arg_expr(args):
            return ", ".join(f"{k}={str(v)}" for k, v in itertools.islice(args.items(), 1, None))

        for tr in traces:
            val = tr.ret_val
            if isinstance(val, np.ndarray):
                tp = ReturnTypeInfo(
                    val.dtype.type, is_array=True, ndim=val.ndim
                )
            else:
                tp = ReturnTypeInfo(type(val))

            if has_last:
                if last_tp == tp:
                    continue
                elif last_tp.is_array and tp.is_array:
                    if last_tp.ndim == tp.ndim:
                        assert last_tp.value_type != tp.value_type
                        if issubclass(last_tp.value_type, numbers.Integral) and issubclass(tp.value_type, numbers.Integral):
                            last_tp.value_type = numbers.Integral
                        elif issubclass(last_tp.value_type, numbers.Real) and issubclass(tp.value_type, numbers.Real):
                            last_tp.value_type = numbers.Real
                        else:
                            if not was_dtype_logged:
                                # Log varying return types with their arguments
                                args0 = get_arg_expr(last_args)
                                args1 = get_arg_expr(tr.arg_vals)
                                msg0 = f"{last_tp.value_type.__name__} for {args0}"
                                msg1 = f"{tp.value_type.__name__} for {args1}"
                                _logger.info(f"varying array types returned from {self.fqname}:{msg0} and {msg1}")
                                was_dtype_logged = True

                            last_tp.value_type = object

                    else:
                        if not was_ndim_logged:
                            args0 = get_arg_expr(last_args)
                            args1 = get_arg_expr(tr.arg_vals)
                            msg0 = f"{last_tp.ndim} for {args0}"
                            msg1 = f"{tp.ndim} for {args1}"
                            _logger.info(f"varying array dimensions returned from {self.fqname}:{msg0} and {msg1}")
                            was_ndim_logged = True

                        last_tp = ReturnTypeInfo(object)

                elif not last_tp.is_array and not tp.is_array:

                    if issubclass(last_tp.value_type, numbers.Integral) and issubclass(tp.value_type, numbers.Integral):
                        last_tp.value_type = numbers.Integral
                    elif issubclass(last_tp.value_type, numbers.Real) and issubclass(tp.value_type, numbers.Real):
                        last_tp.value_type = numbers.Real
                    else:
                        if not was_vtype_logged:
                            args0 = get_arg_expr(last_args)
                            args1 = get_arg_expr(tr.arg_vals)
                            msg0 = f"{last_tp.value_type.__name__} for {args0}"
                            msg1 = f"{last_tp.value_type.__name__} for {args1}"
                            _logger.info(f"varying types returned from {self.fqname}:{msg0} and {msg1}")
                            was_vtype_logged = True
                        last_tp.value_type = object

            else:
                last_tp = tp
                last_args = tr.arg_vals
                has_last = True

        return last_tp


class RuntimeValueInfo:

    def __init__(self, value, mx_class=''):
        self.type_ = type(value)
        self.mx_class = mx_class

    @classmethod
    def init_mxobj(cls, value, module):
        if (value.__class__.__module__)[:len(module)] == module:
            return cls(value, mx_class=value.__class__.__module__ + "." + value.__class__.__qualname__)
        else:
            return cls(value)



class RuntimeRefInfo(RuntimeValueInfo):
    pass


class RuntimeParamInfo(RuntimeValueInfo):
    pass





def replace_first_name(dotted_name: str, name: str):
    names = dotted_name.split(".")
    names[0] = name
    return ".".join(names)


class MxCallTracer(CallTracer):
    """Add return_value to CallTrace"""

    def __init__(
        self,
        module: str,
        logger: CallTraceLogger,
        max_typed_dict_size: int,
        code_filter: Optional[CodeFilter] = None,
        sample_rate: Optional[int] = None,
    ) -> None:
        super().__init__(logger, max_typed_dict_size, code_filter, sample_rate)
        self.module = module

    def handle_call(self, frame: FrameType) -> None:
        if self.sample_rate and random.randrange(self.sample_rate) != 0:
            return
        func = self._get_func(frame)
        if func is None:
            return
        code = frame.f_code
        # I can't figure out a way to access the value sent to a generator via
        # send() from a stack frame.
        if frame in self.traces:
            # resuming a generator; we've already seen this frame
            return
        arg_names = code.co_varnames[0: code.co_argcount]
        arg_vals = {}
        for name in arg_names:
            if name in frame.f_locals:
                arg_vals[name] = frame.f_locals[name]

        self.traces[frame] = CallTrace(func, arg_vals)

    def handle_return(self, frame: FrameType, arg: Any) -> None:
        # In the case of a 'return' event, arg contains the return value, or
        # None, if the block returned because of an unhandled exception. We
        # need to distinguish the exceptional case (not a valid return type)
        # from a function returning (or yielding) None. In the latter case, the
        # the last instruction that was executed should always be a return or a
        # yield.
        last_opcode = frame.f_code.co_code[frame.f_lasti]
        trace = self.traces.get(frame)
        if trace is None:
            return
        elif last_opcode == YIELD_VALUE_OPCODE:
            trace.ret_val = arg
        else:
            if (3, 12) <= sys.version_info < (3, 14):
                ret_opcodes = (RETURN_VALUE_OPCODE, RETURN_CONST_OPCODE)
            else:
                ret_opcodes = (RETURN_VALUE_OPCODE,)
            if last_opcode in ret_opcodes:
                trace.ret_val = arg

            # Extract and store local variables
            # code = frame.f_code
            # arg_names = code.co_varnames[0: code.co_argcount]
            # loc_vals = {}
            # for loc in frame.f_locals:
            #     if loc not in arg_names:
            #         loc_vals[loc] = frame.f_locals[loc]
            # trace.loc_vals = loc_vals

            del self.traces[frame]
            self.logger.log(trace)

    def __call__(self, frame: FrameType, event: str, arg: Any) -> "CallTracer":
        code = frame.f_code
        if (
            event not in SUPPORTED_EVENTS
            or code.co_name == "trace_types"
            or self.should_trace
            and not self.should_trace(code)
        ):
            return self

        # Filter by module name here
        func = get_func(frame)
        if not func or not func.__module__.split(".")[0] == self.module:
            return self

        try:
            if event == EVENT_CALL:
                self.handle_call(frame)
            elif event == EVENT_RETURN:
                self.handle_return(frame, arg)
            else:
                _logger.error("Cannot handle event %s", event)
        except Exception:
            _logger.exception("Failed collecting trace")
        return self


class MxCodeFilter:

    def __call__(self, code):
        # Since called many times, check module name and function name only
        if (
            code.co_filename[-len(MX_MODEL_MOD) - 3: -3] == MX_MODEL_MOD
            or code.co_filename[-len(MX_SPACE_MOD) - 3: -3] == MX_SPACE_MOD
        ):
            if code.co_name[:3] == FORMULA_PREF:
                return True
            elif code.co_name == MX_ASSIGN_REFS:
                return True
            elif code.co_name == "__call__":
                return True

        return False


class MxCallTraceLogger(CallTraceLogger):
    """Log and store/print records collected by a CallTracer."""

    def __init__(self, module: str, new_model_name: str = None) -> None:
        super().__init__()
        self.module = module
        self.new_name = new_model_name
        self._traces = {}  # funcname -> [trace]
        self.cells_info = {}  # funcname -> MethodTypeInfo
        self.ref_info = {}
        self.modules = []
        self.param_info = {}  # class name -> {param: RuntimeParamInfo}

    def log(self, trace: CallTrace) -> None:
        """Log a single call trace."""
        self._traces.setdefault(trace.funcname, []).append(trace)

    def flush(self) -> None:
        for k, v in self._traces.items():

            name_split = k.split(".")
            tr0 = v[0]  # First trace
            if name_split[-1] == MX_ASSIGN_REFS:
                # Extract refs
                for name, val in tr0.arg_vals[MX_SELF].__dict__.items():
                    if is_user_defined(name):
                        fqname = ".".join(name_split[:-1] + [name])
                        self.ref_info[fqname] = RuntimeRefInfo.init_mxobj(val, self.module)
            else:
                # Extract cells
                self.cells_info[k] = info = RuntimeCellsInfo(v)
                if info.module not in self.modules:
                    self.modules.append(info.module)

        self._traces.clear()
        self._get_params()

        if self.new_name:
            self._update_model_name()

    def _update_model_name(self):
        """Change the model name stored in members"""

        
        for i, v in enumerate(self.modules):
            self.modules[i] = replace_first_name(v, self.new_name)

        for key in list(self.cells_info.keys()):
            v = self.cells_info.pop(key)
            v.module = replace_first_name(v.module, self.new_name)
            self.cells_info[replace_first_name(key, self.new_name)] = v

        for key in list(self.ref_info.keys()):
            v = self.ref_info.pop(key)
            if v.mx_class:
                v.mx_class = replace_first_name(v.mx_class, self.new_name)
            self.ref_info[replace_first_name(key, self.new_name)] = v

        for key in list(self.param_info.keys()):
            params = self.param_info[key]
            for v in list(params.values()):
                if v.mx_class:
                    v.mx_class = replace_first_name(v.mx_class, self.new_name)
            self.param_info[replace_first_name(key, self.new_name)] = params

    def _get_params(self):
        module = sys.modules[self.module + "." + MX_MODEL_MOD]
        base_cls = getattr(sys.modules[self.module + "." + MX_SYS_MOD], BASE_MODEL)
        model_cls = next(v for v in module.__dict__.values() if isinstance(v, type) and issubclass(v, base_cls))
        model = next(v for v in module.__dict__.values() if isinstance(v, model_cls))

        for space in model._mx_spaces.values():
            self._walk_space(space, {})

    def _walk_space(self, top_space, params):
        for space in top_space._mx_walk():

            if params:
                key = space.__class__.__module__ + "." + space.__class__.__name__
                val = self.param_info.get(key, None)
                if (not val) or len(val) < len(params):
                    self.param_info[key] = params

            if hasattr(space, "__call__"):
                space_name = space.__class__.__name__[len(SPACE_PREF):]
                param_list = getattr(sys.modules[space.__class__.__module__], SPACE_PARAMS + "_" + space_name)

                for k, v in space._mx_itemspaces.items():
                    next_params = params.copy()
                    if isinstance(k, tuple):
                        for name, arg in zip(param_list, k):
                            next_params[name] = RuntimeParamInfo.init_mxobj(arg, self.module)
                    else:
                        next_params[param_list[0]] = RuntimeParamInfo.init_mxobj(k, self.module)

                    self._walk_space(v, next_params)


@contextmanager
def trace_calls(
    module: str,
    logger: CallTraceLogger,
    max_typed_dict_size: int,
    code_filter: Optional[CodeFilter] = None,
    sample_rate: Optional[int] = None,
) -> Iterator[None]:
    """Enable call tracing for a block of code"""
    old_trace = sys.getprofile()
    sys.setprofile(
        MxCallTracer(module, logger, max_typed_dict_size, code_filter, sample_rate)
    )
    try:
        yield
    finally:
        sys.setprofile(old_trace)
        logger.flush()



# Copyright (c) 2023-2025 Fumito Hamamura <fumito.ham@gmail.com>

# This library is free software: you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation version 3.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library.  If not, see <http://www.gnu.org/licenses/>.

from abc import ABC, abstractmethod
from typing import Union, Sequence, Mapping

try:
    from types import NoneType
except ImportError:  # Python -3.9
    NoneType = type(None)

from functools import cached_property
import libcst as cst
import libcst.matchers as m
from libcst.metadata import ParentNodeProvider, ScopeProvider, GlobalScope, ClassScope

from modelx_cython.consts import (
    FORMULA_PREF,
    GLOBAL_PREF,
    SPACE_PREF,
    MX_SELF,
    MX_ASSIGN_REFS,
    is_user_defined,
)


class LexicalBaseMemberInfo:
    module: str
    cls: str
    name: str

    def __init__(self, module, cls, name):
        self.module: str = module
        self.cls: str = cls
        self.name: str = name

    @cached_property
    @abstractmethod
    def fqname(self):
        pass


class LexicalCellsInfo(LexicalBaseMemberInfo):

    params: Sequence[str]

    def __init__(self, module, cls, name, params) -> None:
        super().__init__(module, cls, name)
        self.params: Sequence[str] = params

    def is_special(self):
        return self.name[:2] == self.name[-2:] == "__"

    @cached_property
    def fqname(self):
        pref = "" if self.is_special() else FORMULA_PREF
        result = self.module + "." + self.cls + "." + pref + self.name
        return result


class LexicalRefInfo(LexicalBaseMemberInfo):

    @cached_property
    def fqname(self):
        return self.module + "." + self.cls + "." + self.name


class ParentScopeAddin:

    def get_parent(self, node, level=0):
        while level:
            node = self.get_metadata(ParentNodeProvider, node)
            level -= 1
        return node

    def _get_scope(self, node, level=0):
        return self.get_metadata(ScopeProvider, self.get_parent(node, level=level))

    def is_space_scope(self, node, level=0):
        scope = self._get_scope(node, level)
        return bool(
            isinstance(scope, ClassScope)
            and scope.name[: len(SPACE_PREF)] == SPACE_PREF
            and isinstance(scope.parent, GlobalScope)
        )


class ModuleVisitor(m.MatcherDecoratableVisitor, ParentScopeAddin):
    METADATA_DEPENDENCIES = (ScopeProvider, ParentNodeProvider)

    def __init__(self, module, source):
        super().__init__()
        self.module = module
        self.source = source
        self.cells_info = {}
        self.ref_info = {}  # {class_name: {name: CombinedRefInfo}}
        self.classes = []
        self.spaces = {}  # Parent class name to list of child space names
        self.cimports = []
        self.wrapper = cst.metadata.MetadataWrapper(cst.parse_module(source))
        self.wrapper.visit(self)

    @m.leave(m.ClassDef())
    def collect_classes(self, original_node):
        name = original_node.name.value
        if name[:len(SPACE_PREF)] == SPACE_PREF:
            self.classes.append(name)

    @m.call_if_inside(m.ClassDef())
    @m.call_if_inside(m.FunctionDef(name=cst.Name("__init__")))
    @m.leave(m.SimpleStatementLine())
    def collect_space_info(self, original_node):
        if self.is_space_scope(original_node, level=2):
            # SimpleStatement in IndentedBlock in FunctionDef in IndentedBlock in ClassDef
            node = original_node

            # Retrieve class node
            cls_node = original_node
            for _ in range(4):
                cls_node = self.get_metadata(ParentNodeProvider, cls_node)
            cls_name = cst.ensure_type(cls_node, cst.ClassDef).name.value

            try:
                target = cst.ensure_type(node.body[0], cst.Assign).targets[0].target
            except Exception:
                return

            # Assuming all member assignments in __init__ to names without prefix "_" are child spaces
            # TODO: Rewrite to a robuster condition. Make modelx.export output child space list
            if (target.value.value == MX_SELF and is_user_defined(target.attr.value)):
                self.spaces.setdefault(cls_name, []).append(target.attr.value)

    @m.call_if_inside(m.ClassDef())
    @m.call_if_inside(m.FunctionDef(name=cst.Name(MX_ASSIGN_REFS)))
    @m.leave(m.SimpleStatementLine())
    def collect_refs_info(self, original_node):

        if self.is_space_scope(original_node, level=2):
            # SimpleStatement in IndentedBlock in FunctionDef in IndentedBlock in ClassDef

            # Retrieve class node
            cls_node = original_node
            for _ in range(4):
                cls_node = self.get_metadata(ParentNodeProvider, cls_node)
            cls_name = cst.ensure_type(cls_node, cst.ClassDef).name.value

            try:
                name = cst.ensure_type(
                    cst.ensure_type(original_node.body[0], cst.Assign).targets[0],
                    cst.AssignTarget,
                ).target.attr.value
            except Exception:  # igonore other than assignments, such as 'pass'
                return

            self.ref_info.setdefault(cls_name, {})[name] = LexicalRefInfo(
                self.module,
                cls_name,
                name
            )

    @m.call_if_inside(m.ClassDef())
    @m.visit(m.FunctionDef())
    def collect_methods(self, original_node):

        if self.is_space_scope(original_node):
            cls_name = cst.ensure_type(
                self.get_parent(original_node, level=2),
                cst.ClassDef,
            ).name.value

            if original_node.name.value[: len(FORMULA_PREF)] == FORMULA_PREF:
                # _f_ methods
                pass
            elif original_node.name.value[: len(GLOBAL_PREF)] == GLOBAL_PREF:
                # _mx_ methods
                pass
            elif (
                    original_node.name.value[:2] == "__"
                    and original_node.name.value != "__call__"
            ):
                # Special methods
                pass
            else:
                # cells
                name = original_node.name.value
                params = [
                    p.name.value
                    for p in original_node.params.params
                             + original_node.params.posonly_params
                    if p.name.value != MX_SELF
                ]

                ci = LexicalCellsInfo(
                    module=self.module,
                    cls=cls_name,
                    name=name,
                    params=params
                )
                self.cells_info.setdefault(cls_name, {})[name]  = ci

        return False




cdef class BaseMxObject:
    pass

cdef class BaseParent(BaseMxObject):

    cdef public dict  _mx_spaces
    cdef public BaseParent _parent
    cdef public BaseModel _model
    cdef public str _name

cdef class BaseModel(BaseParent):
    pass

cdef class BaseSpace(BaseParent):

    cdef bint _mx_is_cells_set
    cdef dict _mx_cells

    cdef BaseSpace _space
    cdef dict _mx_itemspaces 
    cdef public list _mx_roots 




# Copyright (c) 2023-2025 Fumito Hamamura <fumito.ham@gmail.com>
import numbers
# This library is free software: you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation version 3.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library.  If not, see <http://www.gnu.org/licenses/>.

from typing import Union, Sequence, Mapping, Dict, Tuple
import logging

try:
    from types import NoneType
except ImportError:  # Python -3.9
    NoneType = type(None)

from functools import cached_property

from modelx_cython.typedefs import get_type_expr
from modelx_cython.config import TransSpec
from modelx_cython.tracer import RuntimeCellsInfo, MxCallTraceLogger
from modelx_cython.parser import ModuleVisitor, LexicalCellsInfo, LexicalRefInfo

from modelx_cython.consts import (
    SPACE_PREF,
    MODULE_PREF
)
from modelx_cython.typedefs import str_to_type, normalize_type

_logger = logging.getLogger(__name__)


class CombinedCellsInfo(LexicalCellsInfo):
    parent: 'ClassInfo'
    _rt: RuntimeCellsInfo
    _spec: dict
    _spec_ret_t: str

    def __init__(self, cls_info, lx_info, rt_info, spec) -> None:
        super().__init__(
            lx_info.module, lx_info.cls, lx_info.name, lx_info.params
        )
        self.parent = cls_info
        self._rt = rt_info
        self._spec = spec
        self._spec_ret_t = spec.get(TransSpec.RET_T, "")

    @cached_property
    def norm_type(self) -> type:
        if self.has_typeinfo():
            if self._spec_ret_t:
                if self._spec_ret_t in str_to_type:
                    return str_to_type[self._spec_ret_t]
                else:
                    raise ValueError(f"invalid value for spec '{TransSpec.RET_T}': {self._spec_ret_t}")
            else:
                return normalize_type(self._rt.ret_type.value_type)
        else:
            return object

    @cached_property
    def is_real_value(self):
        assert self.has_typeinfo()
        return issubclass(self.norm_type, numbers.Real)

    @cached_property
    def is_array_returned(self):
        assert self.has_typeinfo()
        return self._rt.ret_type.is_array

    def has_typeinfo(self):
        return bool(self._rt)

    def has_args(self):
        return bool(self.params)

    def get_argtype_expr(self, arg: str, c_style=False) -> str:
        if self.has_typeinfo():
            assert arg in self._rt.arg_types
            return get_type_expr(self._rt.arg_types[arg], c_style=c_style)
        else:
            return "object"

    def get_rettype_expr(self, c_style=False):

        if self.has_typeinfo():
            typ = get_type_expr(self.norm_type, c_style=c_style)
            if self.is_real_value and self.is_array_returned:
                return typ + "[" + ", ".join(":" * self._rt.ret_type.ndim) + "]"
            else:
                return typ
        else:
            return "object"

    def is_arg_int(self, arg: str):
        assert self.has_args() and self.has_typeinfo()
        return issubclass(self._rt.arg_types[arg], numbers.Integral)

    @cached_property
    def is_int_args(self):
        assert self.has_args() and self.has_typeinfo()
        for p in self.params:
            if self.is_arg_int(p):
                continue
            else:
                return False
        return True

    def is_arrayable(self):
        assert self.has_args() and self.has_typeinfo()
        if self.is_int_args and self.is_real_value and not self.is_array_returned:
            return True
        else:
            return False

    def get_array_decl_expr(self, rettype_expr="", c_style=False):
        assert self.is_arrayable()
        if not rettype_expr:
            rettype_expr = self.get_rettype_expr(c_style=c_style)

        sizes = self.parent.cells_arg_sizes[tuple(self.params)]
        return rettype_expr + "".join([f"[{str(i)}]" for i in sizes])


class CombinedRefInfo:
    module: str
    cls: str
    name: str
    type_: type = None
    mx_class: str = ''
    decl_type_expr: str = ''
    is_relative: bool = False

    def __init__(self, module,
                 cls,
                 name,
                 rt_info):
        if rt_info:

            if rt_info.mx_class:
                if rt_info.mx_class[:len(module)] == module:
                    # Defined in a child space
                    decl_type_expr = rt_info.mx_class[len(module) + 1:]
                    is_relative = True
                else:
                    decl_type_expr = rt_info.mx_class
                    is_relative = False
            else:
                decl_type_expr = ''
                is_relative = False

            self.module = module
            self.cls = cls
            self.name = name
            self.type_ = rt_info.type_
            self.mx_class = rt_info.mx_class
            self.decl_type_expr = decl_type_expr
            self.is_relative = is_relative
        else:
            self.module = module
            self.cls = cls
            self.name = name

    def get_type_expr(self, c_style=False):
        if self.decl_type_expr:
            return self.decl_type_expr
        else:
            return get_type_expr(self.type_, c_style=c_style)


class ClassInfo:

    name: str
    module: 'ModuleInfo'
    visitor: ModuleVisitor
    logger: MxCallTraceLogger
    cells:  dict  # name -> CombinedCellsInfo
    refs:   dict  # name -> CombinedRefInfo
    spaces: list
    params: dict  # name -> CombinedRefInfo
    _cells_max_args: Dict[Tuple[str], Tuple[int]]
    _max_arg_cells: Dict[Tuple[str], Dict[str, str]]  # {(arg,) : {arg: fqname}}

    def __init__(self, name, module):
        self.name = name
        self.module = module
        self.visitor = module.visitor
        self.logger = module.logger
        self.cells = {}
        self._cells_max_args = {}
        self._max_arg_cells = {}    # keep cells fqname for logging
        self.refs = {}
        self.spaces = []
        self._init_cells()
        self._init_spaces()
        self._init_refs()
        self._add_space_params()

    def _init_cells(self):
        for name, lx_info in self.visitor.cells_info[self.name].items():
            rt_info = self.logger.cells_info.get(lx_info.fqname, None)
            self.cells[name] = CombinedCellsInfo(
                self,
                lx_info, rt_info,
                self.module.spec.get_spec(self.fqname).get(TransSpec.CELLS, {}).get(name, {})
            )
            if rt_info:
                args = tuple(rt_info.max_args)
                maxes = tuple(rt_info.max_args.values())
                if args not in self._cells_max_args:
                    self._cells_max_args[args] = maxes
                    self._max_arg_cells[args] = {k: lx_info.fqname for k in args}
                else:
                    d = dict(zip(args, self._cells_max_args[args]))
                    for k, v in rt_info.max_args.items():
                        if v > d[k]:
                            d[k] = v
                            self._max_arg_cells[args][k] = lx_info.fqname

    def _init_spaces(self):
        self.spaces.extend(self.visitor.spaces.get(self.name, []))

    def _init_refs(self):
        for name, lx_info in self.visitor.ref_info.get(self.name, {}).items():
            rt_info = self.logger.ref_info.get(
                lx_info.fqname, None
            )
            self.refs[name] = CombinedRefInfo(
                self.module.fqname,
                self.name,
                name,
                rt_info=rt_info
            )

    def _add_space_params(self):
        params = self.logger.param_info.get(self.fqname, None)
        if params:
            for param, rt_info in params.items():
                self.refs[param] = CombinedRefInfo(
                    module=self.module.fqname,
                    cls=self.name,
                    name=param,
                    rt_info=rt_info
                )

    @cached_property
    def fqname(self):
        return self.module.fqname + "." + self.name

    @cached_property
    def cells_arg_sizes(self) -> Mapping[Tuple[str], Tuple[int]]:
        # params = self.module.spec.get_spec(self.fqname).get(TransSpec.CELLS_PARAMS, {})

        sizes = {}
        if TransSpec.CELLS_PARAM_SIZE in self.module.spec.get_spec(self.fqname):
            params = self.module.spec.get_spec(self.fqname)[TransSpec.CELLS_PARAM_SIZE]
            # Tuplize 1-arg
            for k, v in params.items():
                if isinstance(k, tuple):
                    sizes[k] = v
                else:
                    sizes[(k,)] = (v,)

        elif TransSpec.CELLS_PARAMS in self.module.spec.get_spec(self.fqname):  # deprecated
            params = self.module.spec.get_spec(self.fqname)[TransSpec.CELLS_PARAMS]
            # Tuplize 1-arg
            for k, v in params.items():
                if TransSpec.SIZE in v:
                    if isinstance(k, tuple):
                        sizes[k] = v[TransSpec.SIZE]
                    else:
                        sizes[(k,)] = (v[TransSpec.SIZE],)

        for args, maxes in self._cells_max_args.items():
            if args in sizes:
                d = dict(zip(args, sizes[args]))
                for k, v in zip(args, maxes):
                    if v + 1 > d[k]:
                        _logger.info(f"Specified max size of {d[k]} for cells parameter {k} in {self.name} is replaced by {v + 1} from {self._max_arg_cells[args][k]}")
                        d[k] = v + 1
                sizes[args] = tuple(d.values())
            else:
                sizes[args] = tuple(i + 1 for i in maxes)

        return sizes


class ModuleInfo:

    fqname: str
    visitor: ModuleVisitor
    logger: MxCallTraceLogger
    spec: TransSpec
    classes: dict   # class name -> ClassInfo

    def __init__(self, fqname: str, visitor: ModuleVisitor, logger: MxCallTraceLogger,
                 spec: TransSpec):

        self.fqname = fqname
        self.visitor = visitor
        self.logger = logger
        self.spec = spec
        self.classes = {}
        self._init_classes()

    def _init_classes(self):
        for c in self.visitor.classes:
            self.classes[c] = ClassInfo(c, self)

    @cached_property
    def cimports(self):
        result = []
        for cls in self.classes.values():
            for r in cls.refs.values():
                if r.decl_type_expr and not r.is_relative:
                    mod = ".".join(r.decl_type_expr.split(".")[:-1])
                    if mod not in result:
                        result.append(mod)
        return result

    @cached_property
    def sub_modules(self):
        result = []
        for cls in self.visitor.spaces:
            result.append(MODULE_PREF + cls[len(SPACE_PREF):])    # replace _c_ with _m_
        return result





# Copyright (c) 2023-2025 Fumito Hamamura <fumito.ham@gmail.com>

# This library is free software: you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation version 3.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library.  If not, see <http://www.gnu.org/licenses/>.

from typing import Union, Sequence, Mapping
try:
    from types import NoneType
except ImportError: # Python -3.9
    NoneType = type(None)

import textwrap
from functools import cached_property
import libcst as cst
from libcst._flatten_sentinel import FlattenSentinel
from libcst._nodes.module import Module
from libcst._nodes.statement import BaseStatement, ClassDef
from libcst._removal_sentinel import RemovalSentinel
import libcst.matchers as m
from libcst.metadata import ParentNodeProvider, ScopeProvider, GlobalScope, ClassScope

from modelx_cython.parser import ParentScopeAddin
from modelx_cython.builder import ModuleInfo, CombinedCellsInfo

from modelx_cython.consts import (
    FORMULA_PREF,
    CY_MOD,
    GLOBAL_PREF,
    VAR_PREF,
    HAS_PREF,
    SPACE_PREF,
    MODULE_PREF,
    MX_SELF,
    MX_SYS_MOD,
    MX_SPACE_MOD,
    MX_ASSIGN_REFS,
    MX_COPY_REFS,
    is_user_defined,
)

from modelx_cython.typedefs import CY_BOOL_T

class PXDGenerator:

    pxd_template = textwrap.dedent("""\
    {cmodule_imports}
    from {package} cimport {MX_SYS_MOD}
    {child_cimports}

    {class_defs}
    """)

    cls_template = textwrap.dedent("""\
    cdef class {class_name}({MX_SYS_MOD}.BaseSpace):

    {private_var_defs}
    {public_var_defs}

        cpdef {MX_COPY_REFS}({class_name} self, object base, object base_root)

    {private_meth_defs}
    {public_meth_defs}
    """)

    def __init__(self, module: ModuleInfo):
        self.module = module

    @cached_property
    def package(self) -> str:
        return self.module.fqname.split(".")[0]

    @cached_property
    def cmodule_imports(self) -> str:
        stmts = []
        for ci in self.module.cimports:
            stmts.append(f"cimport {ci}")
        return "\n".join(stmts)

    @cached_property
    def code(self):
        return self.pxd_template.format(
            package=self.package,
            cmodule_imports=self.cmodule_imports,
            MX_SYS_MOD=MX_SYS_MOD,
            child_cimports=self.child_cimports,
            class_defs=self.class_defs
        )

    @cached_property
    def class_defs(self):
        stmts = []
        for cls in self.module.classes.keys():
            stmts.append(self.a_class_def(cls))
        return "\n\n".join(stmts)

    def a_class_def(self, name):
        return self.cls_template.format(
            class_name=name,
            MX_SYS_MOD=MX_SYS_MOD,
            private_var_defs=textwrap.indent(self.private_var_defs(name), ' ' * 4),
            public_var_defs=textwrap.indent(self.public_var_defs(name), ' ' * 4),
            MX_COPY_REFS=MX_COPY_REFS,
            private_meth_defs=textwrap.indent(self.private_meth_defs(name), ' ' * 4),
            public_meth_defs=textwrap.indent(self.public_meth_defs(name), ' ' * 4)
        )

    @cached_property
    def child_cimports(self):
        # cimports for child spaces
        parent = ".".join(self.module.fqname.split(".")[:-1])
        stmts = []
        for child in self.module.sub_modules:
            stmts.append(f"from cython.cimports.{parent} import {child}\n")

        return "".join(stmts)

    def private_var_defs(self, cls_name):

        cls_info = self.module.classes[cls_name]
        decl_stmts = []
        for cells in cls_info.cells.values():

            assert cells.module == self.module.fqname
            assert cells.cls == cls_name

            if cells.is_special():
                continue

            if cells.has_args():
                if cells.has_typeinfo() and cells.is_arrayable():

                    var_name = VAR_PREF + cells.name
                    var_type = cells.get_array_decl_expr(c_style=True)
                    decl_stmts.append(f"cdef {var_type} {var_name}\n")

                    has_name = HAS_PREF + cells.name
                    has_type = cells.get_array_decl_expr(
                                rettype_expr=CY_BOOL_T, c_style=True)
                    decl_stmts.append(f"cdef {has_type} {has_name}\n")

                else:
                    decl_stmts.append(f"cdef dict {VAR_PREF + cells.name}\n")
            else:
                rettype = cells.get_rettype_expr(c_style=True)
                decl_stmts.append(f"cdef {rettype} {VAR_PREF + cells.name}\n")
                decl_stmts.append(f"cdef {CY_BOOL_T} {HAS_PREF + cells.name}\n")

        return "".join(decl_stmts)

    def public_var_defs(self, cls_name):

        decl_stmts = []
        for ref in self.module.classes[cls_name].refs.values():

            assert ref.module == self.module.fqname and ref.cls == cls_name

            stmt = f"cdef public {ref.get_type_expr(c_style=True)} {ref.name}\n"
            decl_stmts.append(stmt)

        # Declare child spaces
        for space in self.module.classes[cls_name].spaces:
            mod_name = MODULE_PREF + cls_name[len(SPACE_PREF):]  # Replace prefix for submodule
            rel_path = mod_name + "." + MX_SPACE_MOD + "." + SPACE_PREF + space

            stmt = f"cdef public {rel_path} {space}\n"
            decl_stmts.append(stmt)

        return "".join(decl_stmts)

    def _add_param_type_hints(
        self, cls_name: str, cells_name: str
    ) -> str:

        cells = self.module.classes[cls_name].cells[cells_name]
        params = [f"{cls_name} {MX_SELF}"]  # add self first

        # Add parameter type hints
        if cells and cells.has_typeinfo() and cells.has_args():
            for param in cells.params:
                type_ = cells.get_argtype_expr(param, c_style=True)
                params.append(f"{type_} {param}")
        else:
            for p in cells.params:
                params.append(f"object {p}")

        return ", ".join(params)

    def private_meth_defs(self, cls_name):

        decl_stmts = []
        for cells in self.module.classes[cls_name].cells.values():

            if cells.is_special():
                continue

            if cells and cells.has_typeinfo():
                rettype = cells.get_rettype_expr(c_style=True)
                parameters = self._add_param_type_hints(
                    cls_name=cls_name, cells_name=cells.name
                )
                decl_stmts.append(
                    f"cdef {rettype} {FORMULA_PREF + cells.name}({parameters})\n"
                )
            else:
                parameters = self._add_param_type_hints(
                    cls_name=cls_name, cells_name=cells.name
                )
                decl_stmts.append(
                    f"cdef object {FORMULA_PREF + cells.name}({parameters})\n"
                )

        return "".join(decl_stmts)

    def public_meth_defs(self, cls_name):

        decl_stmts = []
        for cells in self.module.classes[cls_name].cells.values():

            if cells.is_special():
                continue

            if cells and cells.has_typeinfo():
                rettype = cells.get_rettype_expr(c_style=True)
                parameters = self._add_param_type_hints(
                    cls_name=cls_name, cells_name=cells.name
                )
                decl_stmts.append(
                    f"cpdef {rettype} {cells.name}({parameters})\n"
                )
            else:
                parameters = self._add_param_type_hints(
                    cls_name=cls_name, cells_name=cells.name
                )
                decl_stmts.append(
                    f"cpdef object {cells.name}({parameters})\n"
                )

        return "".join(decl_stmts)


class ModuleTransformer(m.MatcherDecoratableTransformer, ParentScopeAddin):
    METADATA_DEPENDENCIES = (ScopeProvider, ParentNodeProvider)

    def __init__(
        self,
        source: str,
        module: ModuleInfo
    ) -> None:
        super().__init__()
        self.wrapper = cst.metadata.MetadataWrapper(cst.parse_module(source))
        self._module_node = self.wrapper.module
        self.module = module
        self.package = module.fqname.split(".")[0]

    @property   # cannot use cached_property in Transformer
    def transformed(self):
        return self.wrapper.visit(self)

    def leave_Module(self, original_node: Module, updated_node: Module) -> Module:

        # cimports for child spaces
        parent = ".".join(self.module.fqname.split(".")[:-1])
        stmts = []
        for child in self.module.sub_modules:
            stmts.append(cst.parse_statement(
                f"from cython.cimports.{parent} import {child}",
                config=updated_node.config_for_parsing,
            ))

        return updated_node.with_changes(
            body=(
                cst.parse_statement(
                    f"from cython.cimports.{self.package} import {MX_SYS_MOD}",
                    config=updated_node.config_for_parsing,
                ),
                *stmts,
                cst.parse_statement(
                    f"import cython as {CY_MOD}", config=updated_node.config_for_parsing
                ),
                *updated_node.body,
            )
        )

    def leave_ClassDef(
        self, original_node: ClassDef, updated_node: ClassDef
    ) -> Union[BaseStatement, FlattenSentinel[BaseStatement], RemovalSentinel]:
        cls_name: str = original_node.name.value
        if cls_name[: len(SPACE_PREF)] == SPACE_PREF and isinstance(
            self.get_metadata(ScopeProvider, original_node), GlobalScope
        ):
            cls_info = self.module.classes[cls_name]
            decl_stmts = []
            for cells in cls_info.cells.values():

                assert cells.module == self.module.fqname
                assert cells.cls == cls_name

                if cells.is_special():
                    continue

                if cells.has_args():
                    if cells.has_typeinfo() and cells.is_arrayable():
                        decl_stmts.append(
                            cst.parse_statement(
                                VAR_PREF
                                + cells.name
                                + ": "
                                + cells.get_array_decl_expr(),
                                config=self._module_node.config_for_parsing,
                            )
                        )
                        decl_stmts.append(
                            cst.parse_statement(
                                HAS_PREF
                                + cells.name
                                + ": "
                                + cells.get_array_decl_expr(
                                    rettype_expr=f"{CY_MOD}.{CY_BOOL_T}",
                                ),
                                config=self._module_node.config_for_parsing,
                            )
                        )
                    else:
                        decl_stmts.append(
                            cst.parse_statement(
                                VAR_PREF + cells.name + ": dict",
                                config=self._module_node.config_for_parsing,
                            )
                        )
                else:
                    rettype = cells.get_rettype_expr()
                    decl_stmts.append(
                        cst.parse_statement(
                            VAR_PREF + cells.name + ": " + rettype,
                            config=self._module_node.config_for_parsing,
                        )
                    )

                    decl_stmts.append(
                        cst.parse_statement(
                            HAS_PREF + cells.name + ": " + CY_MOD + "." + CY_BOOL_T,
                            config=self._module_node.config_for_parsing,
                        )
                    )

            is_first = True
            for ref in self.module.classes[cls_name].refs.values():

                assert ref.module == self.module.fqname and ref.cls == cls_name

                stmt = cst.parse_statement(
                    f"{ref.name}: {ref.get_type_expr()}",
                    config=self._module_node.config_for_parsing,
                )
                if is_first:
                    stmt = stmt.with_changes(
                        leading_lines=tuple(stmt.leading_lines) + (cst.EmptyLine(),)
                    )
                    is_first = False

                decl_stmts.append(stmt)

            # Declare child spaces
            is_first = True
            for space in self.module.classes[cls_name].spaces:

                mod_name = MODULE_PREF + cls_name[len(SPACE_PREF):]  # Replace prefix for submodule
                rel_path = mod_name + "." + MX_SPACE_MOD + "." + SPACE_PREF + space

                stmt = cst.parse_statement(
                    f"{space}: {rel_path}",
                    config=self._module_node.config_for_parsing,
                )
                if is_first:
                    stmt = stmt.with_changes(
                        leading_lines=tuple(stmt.leading_lines) + (cst.EmptyLine(),)
                    )
                    is_first = False

                decl_stmts.append(stmt)

            decorator = cst.Decorator(
                decorator=cst.Attribute(value=cst.Name(CY_MOD), attr=cst.Name("cclass"))
            )

            if decl_stmts:
                # Add blank lines below classdef
                decl_stmts[0] = decl_stmts[0].with_changes(
                    leading_lines=tuple(decl_stmts[0].leading_lines) + (cst.EmptyLine(),)
                )
                indented_block = cst.ensure_type(
                    updated_node.body, cst.IndentedBlock
                ).with_changes(body=tuple(decl_stmts) + updated_node.body.body)
                return updated_node.with_changes(
                    decorators=(decorator,), body=indented_block
                )
            else:
                return updated_node.with_changes(decorators=(decorator,))
        else:
            return updated_node

    @m.call_if_inside(m.ClassDef())
    @m.call_if_inside(m.FunctionDef(name=cst.Name("__init__")))
    @m.leave(m.SimpleStatementLine())
    def remove_cache_assigns(self, original_node, updated_node):
        funcdef = self.get_parent(original_node, level=2)
        clsdef = self.get_parent(funcdef, level=2)
        if (
            m.matches(funcdef, m.FunctionDef(name=cst.Name("__init__")))
            and m.matches(clsdef, m.ClassDef())
            and clsdef.name.value[: len(SPACE_PREF)] == SPACE_PREF
            and m.matches(original_node, m.SimpleStatementLine(body=(m.Assign(),)))
            and m.matches(
                original_node.body[0],
                m.Assign(
                    targets=(m.AssignTarget(target=m.Attribute(value=m.Name(MX_SELF))),)
                ),
            )
            and (
                original_node.body[0].targets[0].target.attr.value[: len(HAS_PREF)]
                == HAS_PREF
                or original_node.body[0].targets[0].target.attr.value[: len(VAR_PREF)]
                == VAR_PREF
            )
        ):
            return cst.RemoveFromParent()

        return updated_node

    def _add_param_type_hints(
        self, funcdef: cst.FunctionDef, cls_name: str
    ) -> Union[cst.Parameters, NoneType]:
        param_list = list(funcdef.params.params + funcdef.params.posonly_params)[
            1:
        ]  # remove self

        name = funcdef.name.value
        if name[:len(FORMULA_PREF)] == FORMULA_PREF:
            name = name[len(FORMULA_PREF):]

        cells = self.module.classes[cls_name].cells[name]

        # Add parameter type hints
        updated_params = [funcdef.params.params[0]]  # add self first
        for param in param_list:
            param_name = param.name.value
            if cells.get_argtype_expr(param_name):
                updated_params.append(
                    param.with_changes(
                        annotation=cst.Annotation(
                            annotation=cst.parse_expression(
                                cells.get_argtype_expr(param_name),
                                config=self._module_node.config_for_parsing,
                            )
                        )
                    )
                )
            else:
                updated_params.append(param)

        return funcdef.params.with_changes(params=tuple(updated_params))

    @m.call_if_inside(m.ClassDef())
    @m.call_if_inside(m.FunctionDef(name=cst.Name(MX_COPY_REFS)))
    @m.call_if_inside(m.SimpleStatementLine())
    @m.call_if_inside(m.Assign())
    @m.call_if_inside(m.Attribute())
    @m.leave(m.Name(value="base"))
    def rename_base(self, original_node, updated_node):
        return updated_node.with_changes(value="base_")

    @m.call_if_inside(m.ClassDef())
    @m.leave(m.FunctionDef())
    def update_method(self, original_node, updated_node):

        if self.is_space_scope(original_node):
            cls_name = cst.ensure_type(
                self.get_parent(original_node, level=2),
                cst.ClassDef,
            ).name.value

            meth_name = original_node.name.value
            cls_info = self.module.classes[cls_name]

            if meth_name[: len(FORMULA_PREF)] == FORMULA_PREF:
                # _f_ methods
                cells = cls_info.cells.get(meth_name[len(FORMULA_PREF):])

                decorators = [
                    cst.Decorator(
                        decorator=cst.Attribute(
                            value=cst.Name(CY_MOD), attr=cst.Name("cfunc")
                        )
                    )
                ]
                returns = cst.Annotation(
                    annotation=cst.parse_expression(
                        cells.get_rettype_expr(),
                        config=self._module_node.config_for_parsing,
                    )
                )
                if cells.has_args():
                    parameters = self._add_param_type_hints(
                        updated_node, cls_name=cls_name
                    )
                    return updated_node.with_changes(
                        decorators=decorators, params=parameters, returns=returns
                    )
                else:
                    return updated_node.with_changes(decorators=decorators, returns=returns)


            elif meth_name == MX_COPY_REFS:
                decorators = [
                    cst.Decorator(
                        decorator=cst.Attribute(
                            value=cst.Name(CY_MOD), attr=cst.Name("ccall")
                        )
                    )
                ]

                decl_stmt = cst.parse_statement(
                    f"base_: {cls_name} = {CY_MOD}.cast({cls_name}, base)",
                    config=self._module_node.config_for_parsing,
                )
                decl_stmt = decl_stmt.with_changes(leading_lines=(cst.EmptyLine(),))
                stmts = list(cst.ensure_type(updated_node.body, cst.IndentedBlock).body)
                stmts.insert(0, decl_stmt)
                indented_block = updated_node.body.with_changes(body=stmts)

                return updated_node.with_changes(
                    decorators=decorators, body=indented_block
                )

            elif meth_name[: len(GLOBAL_PREF)] == GLOBAL_PREF:
                # other _mx_ methods
                return updated_node

            elif meth_name == "__call__":
                # Special methods
                cells = cls_info.cells[meth_name]

                if cells and cells.has_typeinfo() and cells.has_args():
                    parameters = self._add_param_type_hints(
                        updated_node, cls_name=cls_name
                    )
                    return updated_node.with_changes(params=parameters)

            elif meth_name[:2] == "__":
                # Special methods
                pass

            else:
                # cells
                cells: CombinedCellsInfo = cls_info.cells[meth_name]

                decorators = [
                    cst.Decorator(
                        decorator=cst.Attribute(
                            value=cst.Name(CY_MOD), attr=cst.Name("ccall")
                        )
                    )
                ]
                # Return type
                returns = cst.Annotation(
                    annotation=cst.parse_expression(
                        cells.get_rettype_expr(),
                        config=self._module_node.config_for_parsing,
                    )
                )
                if cells.has_args():
                    # Add parameter type hints
                    parameters = self._add_param_type_hints(
                        updated_node, cls_name=cls_name
                    )
                    if cells.has_typeinfo() and cells.is_arrayable():

                        # Construct indented_block to replace the original one
                        c_idx_expr = ''.join([f"[{p}]" for p in cells.params])
                        param_expr = f"{', '.join([p for p in cells.params])}"

                        has_expr = f"{MX_SELF}.{HAS_PREF}{meth_name}{c_idx_expr}"
                        v_expr = f"{MX_SELF}.{VAR_PREF}{meth_name}{c_idx_expr}"
                        f_expr = f"{MX_SELF}.{FORMULA_PREF}{meth_name}({param_expr})"

                        args = tuple(cells.params)
                        size = cls_info.cells_arg_sizes[args]

                        idx_range = " and ".join(
                            [f"(0 <= {p} < {i})" for p, i in zip(args, size)])

                        if_stmt = textwrap.dedent(f"""\
                        if {idx_range}:
                            if {has_expr}:
                                return {v_expr}
                            else:
                                val = {f_expr}
                                {v_expr} = val
                                {has_expr} = True
                                return val
                        else:
                            raise IndexError("array index out of range")
                        """)
                        if_node = cst.parse_statement(
                            if_stmt, config=self._module_node.config_for_parsing
                        )
                        indented_block = cst.ensure_type(
                            updated_node.body, cst.IndentedBlock
                        ).with_changes(body=(if_node,))

                        return updated_node.with_changes(
                            decorators=decorators,
                            params=parameters,
                            returns=returns,
                            body=indented_block,
                        )
                    else:
                        return updated_node.with_changes(
                            decorators=decorators,
                            params=parameters,
                            returns=returns,
                            body=self._add_dict_assign(meth_name, updated_node)
                        )
                else:   # No type info, no arg
                    return updated_node.with_changes(
                        decorators=decorators,
                        returns=returns
                    )


        return updated_node

    def _add_dict_assign(self, meth_name: str, updated_node) -> cst.IndentedBlock:
        """Add dict assignment in method

        Example:
            if self._v_meth is None:
                self._v_meth = {}
        """
        init_stmt = cst.parse_statement(
            f"if {MX_SELF}.{VAR_PREF}{meth_name} is None:\n    {MX_SELF}.{VAR_PREF}{meth_name} = {{}}",
            config=self._module_node.config_for_parsing
        )

        return cst.ensure_type(
            updated_node.body, cst.IndentedBlock
        ).with_changes(body=(init_stmt,) + updated_node.body.body)




# Copyright (c) 2023-2025 Fumito Hamamura <fumito.ham@gmail.com>

# This library is free software: you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation version 3.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library.  If not, see <http://www.gnu.org/licenses/>.

from typing import Union, Sequence, Mapping
from modelx_cython.consts import (
    FILE_PREF,
    GLOBAL_PREF,
    MODULE_PREF,
    SPACE_PREF,
)


class TransSpec:

    SPACES = "spaces"
    SPACE_PARAMS = "space_params"
    CELLS = "cells"
    CELLS_PARAM_SIZE = "cells_param_size"
    CELLS_PARAMS = "cells_params"   # deprecated
    SIZE = "size"   # deprecated
    RET_T = "return_type"
    PARAM_T = "param_type"

    def __init__(self, data: dict) -> None:
        
        self._data = data

    def get_spec(self, object_path: str):
        """
        Examples:

            PkgName._mx_classes._c_Projection
            PkgName._m_Parent._m_Child
        """
        def get_subspace(data, name):
            spaces = data.get(self.SPACES)
            if spaces and spaces.get(name):
                return spaces.get(name)
            else:
                return {}
            
        names = object_path.split(".")[1:]  # Remove first
        data = self._data
        while names:
            name = names.pop(0)
            if name[:len(FILE_PREF)] == FILE_PREF:
                continue
            elif name[:len(MODULE_PREF)] == MODULE_PREF:
                name = name[len(MODULE_PREF):]
                data_ = get_subspace(data, name)
                if data_:
                    data = data_
                else:
                    return {}
            elif name[:len(SPACE_PREF)] == SPACE_PREF:
                name = name[len(SPACE_PREF):]
                data_ = get_subspace(data, name)
                if data_:
                    data = data_
                else:
                    return {}

        return data



# Modified from https://github.com/Instagram/MonkeyType/blob/v23.3.0/monkeytype/tracing.py
# See LICENSES/LICENSE_MONKEYTYPE.txt

# Copyright (c) 2017-present, Facebook, Inc.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.
import inspect
import logging
import random
import sys
from abc import ABCMeta, abstractmethod
from contextlib import contextmanager
from types import CodeType, FrameType
from typing import Any, Callable, Dict, Iterator, Optional, Union, cast

import opcode

from monkeytype.compat import cached_property
from monkeytype.typing import get_type
from monkeytype.util import get_func_fqname


logger = logging.getLogger(__name__)


class CallTrace:    # Updated CallTrace
    """CallTrace contains the values observed during a single invocation of a function"""

    def __init__(
        self,
        func: Callable[..., Any],
        arg_vals: Dict[str, Any],
        loc_vals: Optional[Dict[str, Any]] = None,
        ret_val: Optional[Any] = None,
        yield_val: Optional[Any] = None,
    ) -> None:
        """
        Args:
            func: The function where the trace occurred
            arg_vals: The collected argument values
            ret_val: The collected return value. This will be None if the called function returns
                due to an unhandled exception.
            yield_val: The collected yield value. Should not be relevant in modelx
        """
        self.func = func
        self.arg_vals = arg_vals
        self.loc_vals = loc_vals
        self.ret_val = ret_val
        self.yield_val = yield_val

    @property
    def funcname(self) -> str:
        return get_func_fqname(self.func)


class xCallTrace:   # Original CallTrace
    """CallTrace contains the types observed during a single invocation of a function"""

    def __init__(
        self,
        func: Callable[..., Any],
        arg_types: Dict[str, type],
        return_type: Optional[type] = None,
        yield_type: Optional[type] = None,
    ) -> None:
        """
        Args:
            func: The function where the trace occurred
            arg_types: The collected argument types
            return_type: The collected return type. This will be None if the called function returns
                due to an unhandled exception. It will be NoneType if the function returns the value None.
            yield_type: The collected yield type. This will be None if the called function never
                yields. It will be NoneType if the function yields the value None.
        """
        self.func = func
        self.arg_types = arg_types
        self.return_type = return_type
        self.yield_type = yield_type

    def __eq__(self, other: object) -> bool:
        if isinstance(other, self.__class__):
            return self.__dict__ == other.__dict__
        return NotImplemented

    def __repr__(self) -> str:
        return "CallTrace(%s, %s, %s, %s)" % (
            self.func,
            self.arg_types,
            self.return_type,
            self.yield_type,
        )

    def __hash__(self) -> int:
        return hash(
            (
                self.func,
                frozenset(self.arg_types.items()),
                self.return_type,
                self.yield_type,
            )
        )

    def add_yield_type(self, typ: type) -> None:
        if self.yield_type is None:
            self.yield_type = typ
        else:
            self.yield_type = cast(type, Union[self.yield_type, typ])

    @property
    def funcname(self) -> str:
        return get_func_fqname(self.func)


class CallTraceLogger(metaclass=ABCMeta):
    """Log and store/print records collected by a CallTracer."""

    @abstractmethod
    def log(self, trace: CallTrace) -> None:
        """Log a single call trace."""
        pass

    def flush(self) -> None:
        """Flush all logged traces to output / database.

        Not an abstractmethod because it's OK to leave it as a no-op; for very
        simple loggers it may not be necessary to batch-flush traces, and `log`
        can handle everything.
        """
        pass


def get_func_in_mro(obj: Any, code: CodeType) -> Optional[Callable[..., Any]]:
    """Attempt to find a function in a side-effect free way.

    This looks in obj's mro manually and does not invoke any descriptors.
    """
    # FunctionType is incompatible with Callable
    # https://github.com/python/typeshed/issues/1378
    val = inspect.getattr_static(obj, code.co_name, None)
    if val is None:
        return None
    if isinstance(val, (classmethod, staticmethod)):
        cand = val.__func__
    elif isinstance(val, property) and (val.fset is None) and (val.fdel is None):
        cand = cast(Callable[..., Any], val.fget)
    elif cached_property and isinstance(val, cached_property):
        cand = val.func
    else:
        cand = cast(Callable[..., Any], val)
    return _has_code(cand, code)


def _has_code(
    func: Optional[Callable[..., Any]], code: CodeType
) -> Optional[Callable[..., Any]]:
    while func is not None:
        func_code = getattr(func, "__code__", None)
        if func_code is code:
            return func
        # Attempt to find the decorated function
        func = getattr(func, "__wrapped__", None)
    return None


def get_func(frame: FrameType) -> Optional[Callable[..., Any]]:
    """Return the function whose code object corresponds to the supplied stack frame."""
    code = frame.f_code
    if code.co_name is None:
        return None
    # First, try to find the function in globals
    cand = frame.f_globals.get(code.co_name, None)
    func = _has_code(cand, code)
    # If that failed, as will be the case with class and instance methods, try
    # to look up the function from the first argument. In the case of class/instance
    # methods, this should be the class (or an instance of the class) on which our
    # method is defined.
    if func is None and code.co_argcount >= 1:
        first_arg = frame.f_locals.get(code.co_varnames[0])
        func = get_func_in_mro(first_arg, code)
    # If we still can't find the function, as will be the case with static methods,
    # try looking at classes in global scope.
    if func is None:
        for v in frame.f_globals.values():
            if not isinstance(v, type):
                continue
            func = get_func_in_mro(v, code)
            if func is not None:
                break
    return func


RETURN_VALUE_OPCODE = opcode.opmap["RETURN_VALUE"]
YIELD_VALUE_OPCODE = opcode.opmap["YIELD_VALUE"]

# A CodeFilter is a predicate that decides whether or not a the call for the
# supplied code object should be traced.
CodeFilter = Callable[[CodeType], bool]

EVENT_CALL = "call"
EVENT_RETURN = "return"
SUPPORTED_EVENTS = {EVENT_CALL, EVENT_RETURN}


class CallTracer:
    """CallTracer captures the concrete types involved in a function invocation.

    On a per function call basis, CallTracer will record the types of arguments
    supplied, the type of the function's return value (if any), and the types
    of values yielded by the function (if any). It emits a CallTrace object
    that contains the captured types when the function returns.

    Use it like so:

        sys.setprofile(CallTracer(MyCallLogger()))

    """

    def __init__(
        self,
        logger: CallTraceLogger,
        max_typed_dict_size: int,
        code_filter: Optional[CodeFilter] = None,
        sample_rate: Optional[int] = None,
    ) -> None:
        self.logger = logger
        self.traces: Dict[FrameType, CallTrace] = {}
        self.sample_rate = sample_rate
        self.cache: Dict[CodeType, Optional[Callable[..., Any]]] = {}
        self.should_trace = code_filter
        self.max_typed_dict_size = max_typed_dict_size

    def _get_func(self, frame: FrameType) -> Optional[Callable[..., Any]]:
        code = frame.f_code
        if code not in self.cache:
            self.cache[code] = get_func(frame)
        return self.cache[code]

    def handle_call(self, frame: FrameType) -> None:
        if self.sample_rate and random.randrange(self.sample_rate) != 0:
            return
        func = self._get_func(frame)
        if func is None:
            return
        code = frame.f_code
        # I can't figure out a way to access the value sent to a generator via
        # send() from a stack frame.
        if frame in self.traces:
            # resuming a generator; we've already seen this frame
            return
        arg_names = code.co_varnames[0 : code.co_argcount]
        arg_types = {}
        for name in arg_names:
            if name in frame.f_locals:
                arg_types[name] = get_type(
                    frame.f_locals[name], max_typed_dict_size=self.max_typed_dict_size
                )
        self.traces[frame] = CallTrace(func, arg_types)

    def handle_return(self, frame: FrameType, arg: Any) -> None:
        # In the case of a 'return' event, arg contains the return value, or
        # None, if the block returned because of an unhandled exception. We
        # need to distinguish the exceptional case (not a valid return type)
        # from a function returning (or yielding) None. In the latter case, the
        # the last instruction that was executed should always be a return or a
        # yield.
        typ = get_type(arg, max_typed_dict_size=self.max_typed_dict_size)
        last_opcode = frame.f_code.co_code[frame.f_lasti]
        trace = self.traces.get(frame)
        if trace is None:
            return
        elif last_opcode == YIELD_VALUE_OPCODE:
            trace.add_yield_type(typ)
        else:
            if last_opcode == RETURN_VALUE_OPCODE:
                trace.return_type = typ
            del self.traces[frame]
            self.logger.log(trace)

    def __call__(self, frame: FrameType, event: str, arg: Any) -> "CallTracer":
        code = frame.f_code
        if (
            event not in SUPPORTED_EVENTS
            or code.co_name == "trace_types"
            or self.should_trace
            and not self.should_trace(code)
        ):
            return self
        try:
            if event == EVENT_CALL:
                self.handle_call(frame)
            elif event == EVENT_RETURN:
                self.handle_return(frame, arg)
            else:
                logger.error("Cannot handle event %s", event)
        except Exception:
            logger.exception("Failed collecting trace")
        return self


@contextmanager
def trace_calls(
    logger: CallTraceLogger,
    max_typed_dict_size: int,
    code_filter: Optional[CodeFilter] = None,
    sample_rate: Optional[int] = None,
) -> Iterator[None]:
    """Enable call tracing for a block of code"""
    old_trace = sys.getprofile()
    sys.setprofile(CallTracer(logger, max_typed_dict_size, code_filter, sample_rate))
    try:
        yield
    finally:
        sys.setprofile(old_trace)
        logger.flush()



import enum
import numbers
import ctypes

from modelx_cython.consts import CY_MOD

CY_BOOL_T = "bint"
CY_INT_T = "long long"
CY_INT_T_P = "longlong"     # For Pure Python syntax
CY_INT_C_TYPE = ctypes.c_longlong
CY_FLOAT_T = "double"


str_to_type = {
    "bool": bool,
    "int": numbers.Integral,
    "float": numbers.Real,
    "str": str,
    "object": object
}


def normalize_type(typ: type) -> type:

    if issubclass(typ, bool):
        return bool

    if issubclass(typ, numbers.Integral):
        return numbers.Integral

    elif issubclass(typ, numbers.Real):
        return numbers.Real

    elif issubclass(typ, str):
        return str

    else:
        return object


def get_type_expr(typ, c_style=False):

    if issubclass(typ, bool):
        if c_style:
            return CY_BOOL_T
        else:
            return f"{CY_MOD}.{CY_BOOL_T}"
    elif issubclass(typ, numbers.Integral):
        if c_style:
            return CY_INT_T
        else:
            return f"{CY_MOD}.{CY_INT_T_P}"
    elif issubclass(typ, numbers.Real):
        if c_style:
            return CY_FLOAT_T
        else:
            return f"{CY_MOD}.{CY_FLOAT_T}"

    elif issubclass(typ, str):
        return "str"
    else:
        return "object"